# -*- coding: utf-8 -*-

import math
import unicodedata
from XY import XY


def is_zenkaku(char):
    char_width = unicodedata.east_asian_width(char)
    return char_width in u"WFA"


def zenkaku_len(string):
    return len([x for x in string  if is_zenkaku(x)])


def hankaku_len(string):
    return len([x for x in string  if not is_zenkaku(x)])


def string_width(string):
    width = 0
    for c in string:
        char_width = unicodedata.east_asian_width(c)
        if char_width in u"WFA":
            width += 2
        else:
            width += 1

    return width


class TextFolder:
    def __init__(self, box, string, **kwargs):
        self.box = box
        self.string = string
        self.scale = 1
        self.halign = kwargs.get('halign', 'center')
        self.valign = kwargs.get('valign', 'center')
        self.fontsize = kwargs.get('fontsize', 11)
        self.padding = kwargs.get('padding', 12)
        self.lineSpacing = kwargs.get('lineSpacing', 2)

        if kwargs.get('adjustBaseline'):
            self.adjustBaseline = True
        else:
            self.adjustBaseline = False

        self._result = self._lines()

    def textsize(self, string):
        width = zenkaku_len(string) * self.fontsize + \
                hankaku_len(string) * self.fontsize * 0.55
        return (int(math.ceil(width)), self.fontsize)

    def height(self):
        height = 0
        for string in self._result:
            height += self.textsize(string)[1]

        if len(self._result) > 1:
            height += (len(self._result) - 1) * self.lineSpacing

        return height

    def each_line(self):
        size = XY(self.box[2] - self.box[0], self.box[3] - self.box[1])

        if self.valign == 'top':
            height = 0
        elif self.valign == 'bottom':
            height = size.y - self.height()
        else:
            height = int(math.ceil((size.y - self.height()) / 2.0))
        base_xy = XY(self.box[0], self.box[1])

        for string in self._result:
            textsize = self.textsize(string)

            halign = size.x - textsize[0] * self.scale
            if self.halign == 'left':
                x = self.padding
            elif self.halign == 'right':
                x = halign - self.padding
            else:
                x = int(math.ceil(halign / 2.0))

            if self.adjustBaseline:
                height += textsize[1]
            draw_xy = XY(base_xy.x + x, base_xy.y + height)

            yield string, draw_xy

            if self.adjustBaseline:
                height += self.lineSpacing
            else:
                height += textsize[1] + self.lineSpacing

    def outlineBox(self):
        corners = []
        for string, xy in self.each_line():
            textsize = self.textsize(string)
            width = textsize[0] * self.scale
            height = textsize[1] * self.scale

            if self.adjustBaseline:
                xy = XY(xy.x, xy.y - textsize[1])

            corners.append(xy)
            corners.append(XY(xy.x + width, xy.y + height))

        margin = 2  # this is MAGIC number
        box = (min(p.x for p in corners) - margin,
               min(p.y for p in corners) - margin,
               max(p.x for p in corners) + margin,
               max(p.y for p in corners) + margin)

        return box

    def _lines(self):
        lines = []
        size = (self.box[2] - self.box[0], self.box[3] - self.box[1])

        def splitlines(string):
            for line in string.splitlines():
                for subline in line.split("\\n"):
                    yield subline

        height = 0
        truncated = 0
        for line in splitlines(self.string):
            while line:
                string = line.strip()
                for i in range(0, len(string)):
                    length = len(string) - i
                    metrics = self.textsize(string[0:length])

                    if metrics[0] <= size[0]:
                        break

                if size[1] < height + metrics[1]:
                    truncated = 1
                    break

                lines.append(string[0:length])
                line = string[length:]

                height += metrics[1] + self.lineSpacing

        # truncate last line.
        if truncated:
            string = lines.pop()
            for i in range(0, len(string)):
                if i == 0:
                    truncated = string + ' ...'
                else:
                    truncated = string[0:-i] + ' ...'

                metrics = self.textsize(truncated)
                if metrics[0] <= size[0]:
                    lines.append(truncated)
                    break

        return lines



# -*- coding: utf-8 -*-

import math
from TextFolder import TextFolder


class PDFTextFolder(TextFolder):
    def __init__(self, box, string, **kwargs):
        self.canvas = kwargs.get('canvas')
        self.font = kwargs.get('font')
        self.fontsize = kwargs.get('fontsize', 11)

        TextFolder.__init__(self, box, string, **kwargs)

    def textsize(self, string):
        width = self.canvas.stringWidth(string, self.font, self.fontsize)
        return (int(math.ceil(width)), self.fontsize)



# -*- coding: utf-8 -*-


# Basic implementation of namedtuple for 2.1 < Python < 2.6
def namedtuple(name, fields):
    'Only space-delimited fields are supported.'
    def prop(i, name):
        return (name, property(lambda self: self[i]))
    methods = dict(prop(i, f) for i, f in enumerate(fields.split(' ')))
    methods.update({
        '__new__': lambda cls, *args: tuple.__new__(cls, args),
        '__repr__': lambda self: '%s(%s)' % (
            name,
            ', '.join('%s=%r' % (
                f, getattr(self, f)) for f in fields.split(' ')))})
    return type(name, (tuple,), methods)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import re
import sys
import copy
from utils import uuid
from utils.XY import XY
import noderenderer


def unquote(string):
    if string:
        m = re.match('\A(?P<quote>"|\')((.|\s)*)(?P=quote)\Z', string, re.M)
        if m:
            return re.sub("\\\\" + m.group(1), m.group(1), m.group(2))
        else:
            return string
    else:
        return string


class Base(object):
    int_attrs = ['width', 'height']

    def duplicate(self):
        return copy.copy(self)

    def set_attribute(self, attr):
        name = attr.name
        if name in self.int_attrs:
            setattr(self, name, int(attr.value))
        elif hasattr(self, name) and not callable(getattr(self, name)):
            setattr(self, name, unquote(attr.value))
        else:
            class_name = self.__class__.__name__
            msg = "Unknown attribute: %s.%s" % (class_name, attr.name)
            raise AttributeError(msg)

    def set_attributes(self, attrs):
        for attr in attrs:
            self.set_attribute(attr)


class Element(Base):
    basecolor = (255, 255, 255)
    namespace = {}

    @classmethod
    def get(cls, id):
        if not id:
            id = uuid.generate()

        if id not in cls.namespace:
            obj = cls(id)
            cls.namespace[id] = obj

        return cls.namespace[id]

    @classmethod
    def clear(cls):
        cls.namespace = {}

    def __init__(self, id):
        self.id = id
        self.label = ''
        self.xy = XY(0, 0)
        self.group = None
        self.drawable = False
        self.order = 0
        self.color = self.__class__.basecolor
        self.width = 1
        self.height = 1

    def __repr__(self):
        class_name = self.__class__.__name__
        nodeid = self.id
        xy = str(self.xy)
        width = self.width
        height = self.height
        addr = id(self)

        format = "<%(class_name)s '%(nodeid)s' %(xy)s " + \
                 "%(width)dx%(height)d at 0x%(addr)08x>"
        return format % locals()


class DiagramNode(Element):
    basecolor = (255, 255, 255)
    default_shape = 'box'

    @classmethod
    def set_default_shape(cls, shape):
        cls.default_shape = shape

    @classmethod
    def clear(cls):
        cls.namespace = {}
        cls.default_shape = 'box'

    def __init__(self, id):
        super(DiagramNode, self).__init__(id)

        self.label = unquote(id) or ''
        self.shape = self.default_shape
        self.style = None
        self.numbered = None
        self.background = None
        self.description = None
        self.drawable = True

    def set_attributes(self, attrs):
        for attr in attrs:
            value = unquote(attr.value)

            if attr.name == 'style':
                style = value.lower()
                if style in ('solid', 'dotted', 'dashed'):
                    self.style = style
                else:
                    msg = "WARNING: unknown edge style: %s\n" % style
                    sys.stderr.write(msg)
            elif attr.name == 'background':
                if os.path.isfile(value):
                    self.background = value
                else:
                    msg = "WARNING: background image not found: %s\n" % value
                    sys.stderr.write(msg)
            elif attr.name == 'shape':
                try:
                    noderenderer.get(value)
                    self.shape = value
                except:
                    msg = "WARNING: unknown node shape: %s\n" % value
                    sys.stderr.write(msg)
            else:
                self.set_attribute(attr)


class NodeGroup(Element):
    basecolor = (243, 152, 0)

    def __init__(self, id):
        super(NodeGroup, self).__init__(id)

        self.href = None
        self.level = 0
        self.separated = False
        self.nodes = []
        self.edges = []
        self.orientation = 'landscape'

    def duplicate(self):
        copied = Element.duplicate(self)
        copied.nodes = []
        copied.edges = []

        return copied

    def parent(self, level=None):
        if level is None:
            return self.group
        else:
            if self.level < level:
                return None

            group = self
            while group.level != level:
                group = group.group

            return group

    def is_parent(self, other):
        parent = self.parent(other.level)
        return parent == other

    def traverse_nodes(self, preorder=False):
        for node in self.nodes:
            if isinstance(node, NodeGroup):
                if preorder:
                    yield node

                for subnode in node.traverse_nodes(preorder=preorder):
                    yield subnode

                if not preorder:
                    yield node
            else:
                yield node

    def traverse_groups(self, preorder=False):
        for node in self.traverse_nodes(preorder=preorder):
            if isinstance(node, NodeGroup):
                yield node

    def fixiate(self, fixiate_nodes=False):
        if self.separated:
            self.width = 1
            self.height = 1

            return
        elif len(self.nodes) > 0:
            self.width = max(x.xy.x + x.width for x in self.nodes)
            self.height = max(x.xy.y + x.height for x in self.nodes)

        for node in self.nodes:
            if fixiate_nodes:
                node.xy = XY(self.xy.x + node.xy.x,
                             self.xy.y + node.xy.y)

            if isinstance(node, NodeGroup):
                node.fixiate(fixiate_nodes)

    def update_order(self):
        for i, node in enumerate(self.nodes):
            node.order = i

    def set_attributes(self, attrs):
        for attr in attrs:
            value = unquote(attr.value)

            if attr.name == 'orientation':
                orientation = value.lower()
                if orientation in ('landscape', 'portrait'):
                    self.orientation = orientation
                else:
                    msg = "WARNING: unknown diagram orientation: %s\n" % value
                    sys.stderr.write(msg)
            else:
                self.set_attribute(attr)


class Diagram(NodeGroup):
    int_attrs = ['width', 'height', 'fontsize',
                 'node_width', 'node_height', 'span_width', 'span_height']

    def __init__(self):
        super(Diagram, self).__init__(None)

        self.node_width = None
        self.node_height = None
        self.span_width = None
        self.span_height = None
        self.fontsize = None

    def set_attributes(self, attrs):
        for attr in attrs:
            value = unquote(attr.value)

            if attr.name == 'default_shape':
                try:
                    noderenderer.get(value)
                    DiagramNode.set_default_shape(value)
                except:
                    msg = "WARNING: unknown node shape: %s\n" % value
                    sys.stderr.write(msg)
            elif attr.name == 'shape_namespace':
                noderenderer.set_default_namespace(value)
            else:
                self.set_attribute(attr)


class DiagramEdge(Base):
    namespace = {}

    @classmethod
    def get(cls, node1, node2):
        if node1 not in cls.namespace:
            cls.namespace[node1] = {}

        if node2 not in cls.namespace[node1]:
            obj = cls(node1, node2)
            cls.namespace[node1][node2] = obj

        return cls.namespace[node1][node2]

    @classmethod
    def find(cls, node1, node2=None):
        if node1 is None:
            return [e for e in cls.find_all() if e.node2 == node2]
        elif isinstance(node1, NodeGroup):
            edges = cls.find(None, node2)
            return [e for e in edges if e.node1.group.is_parent(node1)]
        else:
            if node1 not in cls.namespace:
                return []

            if node2 is None:
                return cls.namespace[node1].values()

            if node2 not in cls.namespace[node1]:
                return []

        return cls.namespace[node1][node2]

    @classmethod
    def find_all(cls):
        for v1 in cls.namespace.values():
            for v2 in v1.values():
                yield v2

    @classmethod
    def find_by_level(cls, level):
        edges = []
        for e in cls.find_all():
            edge = e.duplicate()
            skips = 0

            if edge.node1.group.level < level:
                skips += 1
            else:
                while edge.node1.group.level != level:
                    edge.node1 = edge.node1.group

            if edge.node2.group.level < level:
                skips += 1
            else:
                while edge.node2.group.level != level:
                    edge.node2 = edge.node2.group

            if skips == 2:
                continue

            edges.append(edge)

        return edges

    @classmethod
    def clear(cls):
        cls.namespace = {}

    def __init__(self, node1, node2):
        self.node1 = node1
        self.node2 = node2
        self.crosspoints = []
        self.skipped = 0

        self.label = None
        self.dir = 'forward'
        self.color = None
        self.style = None
        self.hstyle = None
        self.folded = None

    def __repr__(self):
        class_name = self.__class__.__name__
        node1_id = self.node1.id
        node1_xy = self.node1.xy
        node2_id = self.node2.id
        node2_xy = self.node2.xy
        addr = id(self)

        format = "<%(class_name)s '%(node1_id)s' %(node1_xy)s - " + \
                 "'%(node2_id)s' %(node2_xy)s, at 0x%(addr)08x>"
        return format % locals()

    def set_attributes(self, attrs):
        for attr in attrs:
            value = unquote(attr.value)

            if attr.name == 'dir':
                dir = value.lower()
                if dir in ('back', 'both', 'none', 'forward'):
                    self.dir = dir
                elif dir == '->':
                    self.dir = 'forward'
                elif dir == '<-':
                    self.dir = 'back'
                elif dir == '<->':
                    self.dir = 'both'
                elif dir == '--':
                    self.dir = 'none'
                else:
                    msg = "WARNING: unknown edge dir: %s\n" % dir
                    sys.stderr.write(msg)
            elif attr.name == 'style':
                style = value.lower()
                if style in ('solid', 'dotted', 'dashed'):
                    self.style = style
                else:
                    msg = "WARNING: unknown edge style: %s\n" % style
                    sys.stderr.write(msg)
            elif attr.name == 'hstyle':
                hstyle = value.lower()
                if hstyle in ('generalization', 'composition', 'aggregation'):
                    self.hstyle = hstyle
                else:
                    msg = "WARNING: unknown edge hstyle: %s\n" % hstyle
                    sys.stderr.write(msg)
            elif attr.name == 'folded':
                self.folded = True
            elif attr.name == 'nofolded':
                self.folded = False
            else:
                self.set_attribute(attr)



# -*- coding: utf-8 -*-
from blockdiag.noderenderer import NodeShape
from blockdiag.noderenderer import install_renderer
from blockdiag.utils.XY import XY


class LoopIn(NodeShape):
    def __init__(self, node, metrix=None):
        super(LoopIn, self).__init__(node, metrix)

        m = self.metrix.cell(self.node)
        xdiff = self.metrix.nodeWidth / 4
        ydiff = self.metrix.nodeHeight / 4

        textbox = (m.topLeft().x, m.topLeft().y + ydiff,
                   m.bottomRight().x, m.bottomRight().y)

    def render_shape(self, drawer, format, **kwargs):
        outline = kwargs.get('outline')
        fill = kwargs.get('fill')

        m = self.metrix.cell(self.node)
        xdiff = self.metrix.nodeWidth / 4
        ydiff = self.metrix.nodeHeight / 4

        shape = [XY(m.topLeft().x + xdiff, m.topLeft().y),
                 XY(m.topRight().x - xdiff, m.topLeft().y),
                 XY(m.topRight().x, m.topRight().y + ydiff),
                 XY(m.topRight().x, m.bottomRight().y),
                 XY(m.topLeft().x, m.bottomLeft().y),
                 XY(m.topLeft().x, m.topLeft().y + ydiff),
                 XY(m.topLeft().x + xdiff, m.topLeft().y)]

        # draw outline
        if kwargs.get('shadow'):
            shape = self.shift_shadow(shape)
            drawer.polygon(shape, fill=fill, outline=fill,
                           filter='transp-blur')
        elif self.node.background:
            drawer.polygon(shape, fill=self.node.color,
                             outline=self.node.color)
            drawer.loadImage(self.node.background, self.textbox)
            drawer.polygon(shape, fill="none", outline=outline,
                           style=self.node.style)
        else:
            drawer.polygon(shape, fill=self.node.color, outline=outline,
                           style=self.node.style)


def setup(self):
    install_renderer('flowchart.loopin', LoopIn)



# -*- coding: utf-8 -*-
from blockdiag.noderenderer import NodeShape
from blockdiag.noderenderer import install_renderer
from blockdiag.utils.XY import XY


class Box(NodeShape):
    def render_shape(self, drawer, format, **kwargs):
        outline = kwargs.get('outline')
        fill = kwargs.get('fill')

        # draw outline
        box = self.metrix.cell(self.node).box()
        if kwargs.get('shadow'):
            box = self.shift_shadow(box)
            drawer.rectangle(box, fill=fill, outline=fill,
                             filter='transp-blur')
        elif self.node.background:
            drawer.rectangle(box, fill=self.node.color,
                             outline=self.node.color)
            drawer.loadImage(self.node.background, self.textbox)
            drawer.rectangle(box, outline=outline, style=self.node.style)
        else:
            drawer.rectangle(box, fill=self.node.color, outline=outline,
                             style=self.node.style)


def setup(self):
    install_renderer('box', Box)



# -*- coding: utf-8 -*-

import Image
import ImageDraw
import ImageFont
from TextFolder import TextFolder


class PILTextFolder(TextFolder):
    def __init__(self, box, string, **kwargs):
        font = kwargs.get('font')
        if font:
            fontsize = kwargs.get('fontsize', 11)
            self.ttfont = ImageFont.truetype(font, fontsize)
            self.scale = 1
        else:
            self.ttfont = None
            self.scale = kwargs.get('scale', 1)

        image = Image.new('1', (1, 1))
        self.draw = ImageDraw.Draw(image)

        TextFolder.__init__(self, box, string, **kwargs)

    def textsize(self, string):
        return self.draw.textsize(string, font=self.ttfont)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from types import MethodType
from utils.XY import XY
from DiagramMetrix import *
import noderenderer


class Scaler:
    def scale(self, value):
        ratio = self.scale_ratio

        if isinstance(value, XY):
            ret = XY(value.x * ratio, value.y * ratio)
        elif isinstance(value, tuple):
            ret = tuple([self.scale(x) for x in value])
        elif isinstance(value, list):
            ret = [self.scale(x) for x in value]
        elif isinstance(value, int):
            ret = value * ratio
        else:
            ret = value

        return ret

    def define_proxy_attribute(self, name):
        value = getattr(self.metrix, name)
        setattr(self, name, value)

    def define_scale_attribute(self, name):
        if hasattr(self.metrix, name):
            value = getattr(self.metrix, name)
            setattr(self, name, self.scale(value))

    def define_scale_method(self, name):
        def scale_method(self, *args, **kwargs):
            func = getattr(self.metrix, name)
            ret = func(*args, **kwargs)

            return self.scale(ret)

        method = MethodType(scale_method, self, self.__class__)
        setattr(self, name, method)


class PngDiagramMetrix(Scaler):
    def __init__(self, diagram, scale, **kwargs):
        self.scale_ratio = scale
        self.metrix = DiagramMetrix(diagram, **kwargs)

        # variables
        self.define_scale_attribute('cellSize')
        self.define_scale_attribute('nodePadding')
        self.define_scale_attribute('lineSpacing')
        self.define_scale_attribute('shadowOffsetX')
        self.define_scale_attribute('shadowOffsetY')
        self.define_scale_attribute('nodeWidth')
        self.define_scale_attribute('nodeHeight')
        self.define_scale_attribute('spanWidth')
        self.define_scale_attribute('spanHeight')
        self.define_scale_attribute('fontSize')
        self.define_scale_attribute('pageMargin')

        # methods
        self.define_scale_method('pageSize')

    def originalMetrix(self):
        return self.metrix

    def cell(self, node):
        base = NodeMetrix(node, self.originalMetrix())
        return PngNodeMetrix(node, self, base)

    def node(self, node):
        metrix = self.originalMetrix()
        renderer = noderenderer.get(node.shape)

        if hasattr(renderer, 'render'):
            base = renderer(node, metrix)
        else:
            base = NodeMetrix(node, metrix)

        return PngNodeMetrix(node, self, base)

    def group(self, group):
        return PngNodeMetrix(group, self)

    def edge(self, edge):
        return PngEdgeMetrix(edge, self)


class PngNodeMetrix(Scaler):
    def __init__(self, node, metrix, base_metrix):
        self.scale_ratio = metrix.scale_ratio
        self.metrix = base_metrix

        # variables
        self.define_scale_attribute('x')
        self.define_scale_attribute('y')

        # methods
        self.define_scale_method('box')
        self.define_scale_method('marginBox')
        self.define_scale_method('coreBox')
        self.define_scale_method('groupLabelBox')

        self.define_scale_method('topLeft')
        self.define_scale_method('topCenter')
        self.define_scale_method('topRight')
        self.define_scale_method('bottomLeft')
        self.define_scale_method('bottomCenter')
        self.define_scale_method('bottomRight')
        self.define_scale_method('leftCenter')
        self.define_scale_method('rightCenter')
        self.define_scale_method('center')

        self.define_scale_method('top')
        self.define_scale_method('bottom')
        self.define_scale_method('right')
        self.define_scale_method('left')


class PngEdgeMetrix(EdgeMetrix):
    pass
